/*
 * lexer.l: Analizador lxico Flex para 'bla'
 * Alejandro Luque, 2000
 */

%{
#include "parser.tab.c"
#define MAX_INCLUDE_DEPTH 10
YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
int include_lines[MAX_INCLUDE_DEPTH];
char *include_names[MAX_INCLUDE_DEPTH];
int include_sp = 0;
int i;
int strl;
char *strbuf;
extern char *incdir;
%}

WS      [ \t]+
ID      [A-Za-z\\\$_][A-Za-z0-9\\\$\._]*
COMMENT ^#(.*)\n
INUM    [0-9]+     

%x str

%%

{WS}
{COMMENT}        include_lines[include_sp] ++;
\n               include_lines[include_sp] ++;

\"               i = 0; BEGIN(str);

<str>\"          {
                 BEGIN(INITIAL);
		 strbuf[i] = '\0';
		 yyset(strbuf);
		 return STRING;
}
<str>\n          {
                 fprintf(stderr, "Unterminated string.\n");
		 BEGIN(INITIAL);
}

<str>\\[0-7]{1,3} {
                 /* octal escape sequence */
                 int result;

                 (void) sscanf( yytext + 1, "%o", &result );

                 if ( result > 0xff )
                     fprintf(stderr, "Unknown escape secuence.\n");

                 strbuf[i++] = result;
		 if(i >= strl - 1)
		      strbuf = realloc(strbuf, strl += 10);
		 
}

<str>\\[0-9]+    {
                 fprintf(stderr, "Unknown escape secuence.\n");

}

<str>\\n         {
                 strbuf[i++] = '\n';
		 if(i >= strl - 1)
		      strbuf = realloc(strbuf, strl += 10);
}
<str>\\t         {
                 strbuf[i++] = '\t';
		 if(i >= strl - 1)
		      strbuf = realloc(strbuf, strl += 10);
}
<str>\\r         {
                 strbuf[i++] = '\r';
		 if(i >= strl - 1)
		      strbuf = realloc(strbuf, strl += 10);
}
<str>\\b         {
                 strbuf[i++] = '\b';
		 if(i >= strl - 1)
		      strbuf = realloc(strbuf, strl += 10);
}
<str>\\f         {
                 strbuf[i++] = '\f';
		 if(i >= strl - 1)
		      strbuf = realloc(strbuf, strl += 10);
}

<str>\\(.|\n)    {
                 strbuf[i++] = yytext[1];
                 if(yytext[1] == '\n') include_lines[include_sp]++;
                 if(i >= strl - 1)
		      strbuf = realloc(strbuf, strl += 10);
}

<str>[^\\\n\"]+  {
                 char *yptr = yytext;
                 if(i + strlen(yptr) >= strl - 1){
		      strbuf = realloc(strbuf, i + strlen(yptr) + 1);
		      strl = i + strlen(yptr) + 1;
		 }
                 while (*yptr) 
	              strbuf[i++] = *yptr++;
}

include          return INCLUDE;
do               return DO;
                 
--               yylval.val = (char*) NULL; return STRING;

{ID}             yyset(yytext); return ID;
{INUM}           yyset(yytext); return STRING;

[^ \n\t]         return yytext[0];

<<EOF>> {
       if(include_sp == 0){
              yyterminate();
	      free(include_names[include_sp]);
       }
       else {
              yy_delete_buffer(YY_CURRENT_BUFFER);
	      yy_switch_to_buffer(include_stack[include_sp]);
	      free(include_names[include_sp--]);
       }
}

%%

void yyset(char *m)
{
  yylval.val = (char*) malloc(strlen(m) + 1);
  strcpy((char*) yylval.val, m);
}

void yyinclude(char *fname)
{
  char *actname, c = DIR_SEPARATOR;

  if(include_sp > MAX_INCLUDE_DEPTH) {
    fprintf(stderr, "Includes nested too deeply\n");
    exit(1);
  }
  if(incdir == NULL) {
    if(!(yyin = fopen(fname, "r"))) {
      fprintf(stderr, "Unable to open included file %s\n", fname);
      exit(1);
    }
  } else {
    if(incdir[strlen(incdir)] == DIR_SEPARATOR){
      actname = xmalloc(strlen(incdir) + strlen(fname) + 1);
      strcpy(actname, incdir);
    } else {
      actname = xmalloc(strlen(incdir) + strlen(fname) + 2);
      strcpy(actname, incdir);
      strncat(actname, &c, 1);
    }
    strcat(actname, fname);
    if(!(yyin = fopen(actname, "r"))) {
      fprintf(stderr, "Unable to open included file %s\n", fname);
      exit(1);
    }
  }

  include_lines[++include_sp] = 1;
  include_stack[include_sp] = YY_CURRENT_BUFFER;
  include_names[include_sp] = xmalloc(strlen(fname) + 1);
  strcpy(include_names[include_sp], fname);

  yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
  BEGIN(INITIAL);
}

void yyerror(char *msg)
{
  for(i = 0; i < include_sp; i++)
    fprintf(stderr, "In file included from %s:%d:\n", include_names[i],
	    include_lines[i]);
  fprintf(stderr, "%s:%d: %s.\n", 
	  include_names[include_sp], include_lines[include_sp], msg);
}

void yyinit(char *fname)
{
  include_lines[0] = 1;
  include_names[0] = xmalloc(strlen(fname) + 1);
  strcpy(include_names[0], fname);
  strbuf = (char *) xmalloc(20);
  strl = 20;
}
